ARPC = LibStub("AceAddon-3.0"):NewAddon("AngryRaidPoints", "AceConsole-3.0", "AceEvent-3.0", "AceComm-3.0", "AceTimer-3.0")
local LootItem = {}
local AceGUI = LibStub("AceGUI-3.0")
local libS = LibStub("AceSerializer-3.0")
local libC = LibStub("LibCompress")
local libCE = libC:GetAddonEncodeTable()

local default_channel = "RAID"

BINDING_HEADER_ARPC		= "Angry Raid Points Client"
BINDING_NAME_ARPC_TOGGLE = "Toggle Window"

local MINIMUM_MS = 10
local MINIMUM_OS = 1

local STATUS_NOT_BIDDING = 0
local STATUS_OFF_SPEC = 1
local STATUS_MAIN_SPEC = 2

local STATE_NOT_STARTED = 0
local STATE_LAST_CALL = 1
local STATE_BIDDING_MS = 2
local STATE_BIDDING_OS = 3
local STATE_BIDDING_MSTRIAL = 4
local STATE_BIDDING_OSTRIAL = 5
local STATE_FINISHED = 6

local ilvlMap = {
	[522] = "", [528] = "TF", [535] = "H", [541] = "H TF", -- Tier 15 (ToT)
	[553] = "", [559] = "WF", [566] = "H", [572] = "H WF", -- Tier 16 (SoO)
}

local normalLoot = { 522, 553 }
local normalLoot = { 528, 553 }
local heroicLoot = { 535, 566 }
local heroicPlusLoot = { 535, 566 }

-- Format for addon com:
--
-- { "LOOT",
--	 {
--		 { [Item Link], [Quantity Remaining], [Handler or nil] },
--		 ...
--	 }
-- }

-- { "BIDLIST",
--	 [Item Link],
--	 {
--		{ // Member MS tier
--			{ [Player], [Current Points], [Bid, or 0 for passed, or nil for hasn't bid yet] },
--			...
--		},
--		{ // Member OS tier
--			{ [Player], [Current Points], [Bid, or 0 for passed, or nil for hasn't bid yet] },
--			...
--		},
--		{ // Trial MS tier
--			{ [Player], [Current Points], [Bid, or 0 for passed, or nil for hasn't bid yet] },
--			...
--		},
--		{ // Trial OS tier
--			{ [Player], [Current Points], [Bid, or 0 for passed, or nil for hasn't bid yet] },
--			...
--		},
--	 },
--	 [State, 0=bidding not started, 1=last call announced, 2=Member MS tier bidding, 3=Member OS tier bidding, 4=Trial MS tier bidding, 5=Trial OS tier bidding],
--	 [HighBidder (name) or nil],
--	 [CurrentBidder (numeric index) or nil if not started],
--	 [Quantity remaining, copied here from LOOT for convenience],
--	 [Number of item currently being bid, will be 1 if bidding hasn't started yet],
--	 [Minimum bid required for current bidder, or nil if not started]
-- }

-- { "BIDDING", [Item Link], [Current Points, or nil to remove yourself], [MS? true or false] }

-- { "BIDDING_ACK", [Item Link], [Error text if there is an error, or nil] }

-- { "BID", [Item Link], [Bid Amount, or 0 to pass] }

-- { "BID_ACK", [Item Link], [Error text if there is an error, or nil] }

-- { "WINNER", [Item Link], [Number of the one won, 1 if there's only 1], [Player, or nil if DE'ed], [Winning Bid Amount, or nil if DE'ed], [Loot Log Text] }

-- { "REQUEST_STATUS" }

-- { "VER_QUERY" }
-- { "VERSION", [Version], [SVN Revision #] }
-- { "ASSIGN", [Item Link], [Quantity] }
-- { "QUERY" }
-- { "UPDATE" }
-- { "READY" }
-- { "RESET" }

-- Constants to make dealing with the addon com msgs easier
local COMMAND = 1

local LOOT_Items = 2

local LOOT_Items_Item_Link = 1
local LOOT_Items_Quantity = 2
local LOOT_Items_Handler = 3

local BIDLIST_Item_Link = 2
local BIDLIST_Bidders = 3
local BIDLIST_State = 4
local BIDLIST_HighBidder = 5
local BIDLIST_CurrentBidder = 6
local BIDLIST_Quantity = 7
local BIDLIST_Number = 8
local BIDLIST_MinBid = 9

local BIDLIST_Bidders_MS = 1
local BIDLIST_Bidders_OS = 2
local BIDLIST_Bidders_MSTrial = 3
local BIDLIST_Bidders_OSTrial = 4

local BIDLIST_Bidders_Player = 1
local BIDLIST_Bidders_Points = 2
local BIDLIST_Bidders_Bid = 3

local BIDDING_Item_Link = 2
local BIDDING_Points = 3
local BIDDING_MS = 4

local BIDDING_ACK_Item_Link = 2
local BIDDING_ACK_Error = 3

local BID_Item_Link = 2
local BID_Bid = 3

local BID_ACK_Item_Link = 2
local BID_ACK_Error = 3

local WINNER_Item_Link = 2
local WINNER_Number = 3
local WINNER_Player = 4
local WINNER_Bid = 5
local WINNER_Loot_Log_Text = 6

local VERSION_Version = 2
local VERSION_SVN_Revision = 3

local ASSIGN_Item_Link = 2
local ASSIGN_Quantity = 3

local lootList = {}

local function Error(message)
	ARPC:Print(RED_FONT_COLOR_CODE.."Error:|r "..message)
end

function ARPC_Toggle()
	if ARPC.window:IsShown() then 
		ARPC.window:Hide() 
	else 
		ARPC.window:Show() 
	end
end

local function TestItems()
	ARPC:ProcessMessage("Stingray", {
		"LOOT",
		{
			{ "|cffa335ee|Hitem:96685:0:0:0:0:0:0:0:90:0:0|h[Handwraps of the Witch Doctor]|h|r", 1, nil },
			{ "|cffa335ee|Hitem:96808:0:0:0:0:0:0:0:90:0:0|h[Tortos' Discarded Shell]|h|r", 1, nil }
		}
	})
	ARPC:ProcessMessage("Stingray", {
		"LOOT",
		{
			{ "|cffa335ee|Hitem:96685:0:0:0:0:0:0:0:90:0:0|h[Handwraps of the Witch Doctor]|h|r", 1, nil },
			{ "|cffa335ee|Hitem:96808:0:0:0:0:0:0:0:90:0:0|h[Tortos' Discarded Shell]|h|r", 2, "Ermod" },
			{ "|cffa335ee|Hitem:96487:0:0:0:0:0:0:0:90:0:0|h[Ritual Dagger of the Mind's Eye]|h|r", 3, "Ermod" },
			{ "|cffa335ee|Hitem:96537:0:0:0:0:0:0:0:90:0:0|h[Passionfire Choker]|h|r", 1, "Gulvan" },
			{ "|cffa335ee|Hitem:96507:0:0:0:0:0:0:0:90:0:0|h[Stolen Relic of Zuldazar]|h|r", 2, "Yukio" }
		}
	})

	ARPC:ProcessMessage("Stingray", {
		"BIDLIST",
		"|cffa335ee|Hitem:96808:0:0:0:0:0:0:0:90:0:0|h[Tortos' Discarded Shell]|h|r",
		{
			{ { "Ermod", 234, 12 }, { "Stingray", 342, 0 }, { "Yukio", 1234, 15 }, { "Zanrith", 1434, nil } },
			{ { "Nachwarrior", 1534, nil } },
			{},
			{ { "Endule", 9999, nil } }
		},
		STATE_BIDDING_MS,
		"Yukio",
		4,
		3,
		2,
		16
	})
	ARPC:ProcessMessage("Stingray", { "WINNER", "|cffa335ee|Hitem:96808:0:0:0:0:0:0:0:90:0:0|h[Tortos' Discarded Shell]|h|r", 1, "Yukio", 14 })

	ARPC:ProcessMessage("Stingray", {
		"BIDLIST",
		"|cffa335ee|Hitem:96507:0:0:0:0:0:0:0:90:0:0|h[Stolen Relic of Zuldazar]|h|r",
		{
			{ { "Tisma", 230, nil }, { "Ermod", 234, nil } },
			{ { "Klavak", 534, nil }, { "Nachwarrior", 1534, nil } },
			{},
			{}
		},
		STATE_BIDDING_MS,
		nil,
		1,
		2,
		1,
		10
	})
	ARPC:ProcessMessage("Stingray", {
		"BIDLIST",
		"|cffa335ee|Hitem:96537:0:0:0:0:0:0:0:90:0:0|h[Passionfire Choker]|h|r",
		{ {}, {}, {}, {} },
		STATE_LAST_CALL,
		nil,
		nil,
		nil,
		1,
		1,
		nil
	})
	ARPC:ProcessMessage("Stingray", { "WINNER", "|cffa335ee|Hitem:96507:0:0:0:0:0:0:0:90:0:0|h[Stolen Relic of Zuldazar]|h|r", 1, "Ermod", 11 })
end

local function TestItems2()
	ARPC:ProcessMessage("Stingray", { "WINNER", "|cffa335ee|Hitem:96507:0:0:0:0:0:0:0:90:0:0|h[Stolen Relic of Zuldazar]|h|r", 2, nil, nil })
	ARPC:ProcessMessage("Stingray", {
		"LOOT",
		{
			{ "|cffa335ee|Hitem:96685:0:0:0:0:0:0:0:90:0:0|h[Handwraps of the Witch Doctor]|h|r", 1, nil },
			{ "|cffa335ee|Hitem:96808:0:0:0:0:0:0:0:90:0:0|h[Tortos' Discarded Shell]|h|r", 2, "Ermod" },
			{ "|cffa335ee|Hitem:96487:0:0:0:0:0:0:0:90:0:0|h[Ritual Dagger of the Mind's Eye]|h|r", 3, "Ermod" },
			{ "|cffa335ee|Hitem:96537:0:0:0:0:0:0:0:90:0:0|h[Passionfire Choker]|h|r", 1, "Gulvan" }
		}
	})
end

function ARPC:ReceiveMessage(prefix, data, channel, sender)
	if prefix ~= "ARP" then return end
	
	local one = libCE:Decode(data) -- Decode the compressed data
	
	local two, message = libC:Decompress(one) -- Decompress the decoded data
	
	if not two then error("Error decompressing: " .. message); return end
	
	local success, final = libS:Deserialize(two) -- Deserialize the decompressed data
	if not success then error("Error deserializing " .. final); return end

	self:ProcessMessage( sender, final )
end

function ARPC:SendMessage(data, channel, target)
	local one = libS:Serialize( data )
	local two = libC:CompressHuffman(one)
	local final = libCE:Encode(two)
	local destChannel = channel or default_channel

	if destChannel ~= "RAID" or IsInRaid(LE_PARTY_CATEGORY_HOME) then
		self:SendCommMessage("ARP", final, destChannel, target, "NORMAL")
	end
end

function ARPC:ProcessMessage(sender, data)
	-- ARPC:Print("Got "..data[COMMAND].." from "..sender)
	if data[COMMAND] == "LOOT" then
		local count = 0
		local links = {}
		for index, item in pairs( data[LOOT_Items] ) do
			local result = LootItem:add(item)
			if result then count = count + 1 end
			links[ item[LOOT_Items_Item_Link] ] = true
		end
		for link,item in pairs( lootList ) do
			if not links[ link ] then
				item:ItemFinished()
			end
		end
		if count > 0 then
			if not ARPC.window:IsShown() then ARPC.window:Show() end
			LootItem:ClearFinished()
		end
	elseif data[COMMAND] == "BIDLIST" then
		if LootItem:get( data[BIDLIST_Item_Link] ) then
			LootItem:get(data[BIDLIST_Item_Link]):SetBidList(data)
		end
	elseif data[COMMAND] == "BIDDING_ACK" then
		if LootItem:get( data[BIDDING_ACK_Item_Link] ) then
			LootItem:get( data[BIDDING_ACK_Item_Link] ):ReceivedBiddingAck( data[BIDDING_ACK_Error] )
		end
	elseif data[COMMAND] == "BID_ACK" then
		if LootItem:get( data[BID_ACK_Item_Link] ) then
			LootItem:get( data[BID_ACK_Item_Link] ):ReceivedBidAck( data[BID_ACK_Error] )
		end
	elseif data[COMMAND] == "WINNER" then
		if LootItem:get( data[WINNER_Item_Link] ) then
			LootItem:get( data[WINNER_Item_Link] ):ProcessWinner(data[WINNER_Number], data[WINNER_Player], data[WINNER_Bid])
		end
	end
end

local waitingForItemData = {}
function ARPC:GET_ITEM_INFO_RECEIVED()
	for index, item_link in pairs(waitingForItemData) do
		if GetItemInfo(item_link) then
			local loot = lootList[item_link]
			if loot then
				loot:UpdateFrameInitial()
			end
			table.remove(waitingForItemData, index)
		end
	end
end

local function UpdateStatusBar()
	if ARPC_CurrentPoints == nil then
		ARPC.window:SetStatusText("Current Points: Not Set")
	else
		ARPC.window:SetStatusText("Current Points: "..ARPC_CurrentPoints)
	end
end

local function CreateWindow()
	local window = AceGUI:Create("Frame")
	ARPC.window = window
	window:SetTitle("Angry Raid Points")
	window:SetLayout('Fill')
	window:SetWidth(450)
	window:Hide()
	UpdateStatusBar()

	scrollcontainer = AceGUI:Create("SimpleGroup")
	scrollcontainer:SetFullWidth(true)
	scrollcontainer:SetFullHeight(true)
	scrollcontainer:SetLayout("Fill")

	window:AddChild(scrollcontainer)

	scroll = AceGUI:Create("ScrollFrame")
	scroll:SetLayout("List")
	scrollcontainer:AddChild(scroll)
	window.container = scroll

	window:PauseLayout()
	-- ARPC.ShowAll = true
	-- local checkbox = AceGUI:Create("CheckBox")
	-- checkbox:SetLabel("Show All")
	-- checkbox:SetValue(ARPC.ShowAll)
	-- checkbox:ClearAllPoints()
	-- checkbox:SetWidth(90)
	-- checkbox:SetPoint("BOTTOMRIGHT", window.frame, "BOTTOMRIGHT", -132, 15)
	-- checkbox:SetCallback("OnValueChanged", function(widget, event, value)
	-- 	ARPC.ShowAll = value
	-- 	LootItem:UpdateAllVisibilities()
	-- end)
	-- window:AddChild(checkbox)
end

local function CreateItemButton(parent, icon, text, ...)
	local f = CreateFrame("Button", nil, parent)
	f:SetPoint(...)
	f:SetWidth(24)
	f:SetHeight(24)
	f:SetNormalTexture("Interface\\BUTTONS\\UI-GroupLoot-"..icon.."-Up")
	f:SetPushedTexture("Interface\\BUTTONS\\UI-GroupLoot-"..icon.."-Down")
	f:SetHighlightTexture("Interface\\BUTTONS\\UI-GroupLoot-"..icon.."-Highlight")
	f.parent = parent
	f.tiptext = text
	f:SetScript("OnEnter", function(frame)
		GameTooltip:SetOwner(frame, "ANCHOR_RIGHT")
		GameTooltip:SetText(frame.tiptext)
		GameTooltip:Show()
	end)
	f:SetScript("OnLeave", function(frame) GameTooltip:Hide() end)
	local txt = f:CreateFontString(nil, nil, "GameFontHighlightSmallOutline")
	txt:SetPoint("CENTER", 0, 0)

	return f
end

function LootItem:CreateFrame()
	local group = AceGUI:Create("SimpleGroup")
	local frame = group.frame
	group:PauseLayout()
	group:SetFullWidth(true)
	group:SetAutoAdjustHeight(false)
	group:SetHeight(52)
	ARPC.window.container:AddChild(group)

	frame:SetBackdrop({
		bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
		edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
		tile = true, tileSize = 16, edgeSize = 16,
		insets = { left = 3, right = 3, top = 5, bottom = 3 }
	})
	frame:SetBackdropColor(0.1,0.1,0.1,0.5)
	frame:SetBackdropBorderColor(0.4,0.4,0.4)

	local icon = CreateFrame("Button", nil, frame)
	group.icon = icon
	icon.parent = self
	icon:SetScript("OnEnter", function(frame)
		GameTooltip:SetOwner(frame, "ANCHOR_LEFT", -4, -32)
		GameTooltip:SetHyperlink(frame.parent.item_link)
		GameTooltip:Show()
	end)
	icon:SetScript("OnLeave", function(frame) GameTooltip:Hide() end)
	icon:SetWidth(32)
	icon:SetHeight(32)
	icon:SetPoint("TOPLEFT", frame, "TOPLEFT", 10, -10)

	local item_lbl = frame:CreateFontString()
	group.item_lbl = item_lbl
	item_lbl:SetFontObject("GameFontNormalLarge")
	item_lbl:SetJustifyH("LEFT")
	item_lbl:SetPoint("TOPLEFT", frame, "TOPLEFT", 50, -10)
	item_lbl:SetPoint("TOPRIGHT", frame, "TOPRIGHT", -104, -10)

	local status_lbl = frame:CreateFontString()
	status_lbl:SetFontObject("GameFontHighlight")
	status_lbl:SetPoint("TOPLEFT", item_lbl, "BOTTOMLEFT", 0, -5)
	status_lbl:SetPoint("TOPRIGHT", item_lbl, "BOTTOMRIGHT", 0, -5)
	status_lbl:SetJustifyH("LEFT")
	group.status_lbl = status_lbl

	local bidders_lbl = frame:CreateFontString()
	bidders_lbl:SetFontObject("GameFontHighlight")
	bidders_lbl:SetPoint("TOPLEFT", status_lbl, "BOTTOMLEFT", 0, -5)
	bidders_lbl:SetPoint("TOPRIGHT", status_lbl, "BOTTOMRIGHT", -90, -5)
	bidders_lbl:SetJustifyH("LEFT")
	group.bidders_lbl = bidders_lbl

	local bidders2_lbl = frame:CreateFontString()
	bidders2_lbl:SetFontObject("GameFontHighlight")
	bidders2_lbl:SetPoint("TOPLEFT", status_lbl, "BOTTOMRIGHT", -95, -4)
	bidders2_lbl:SetPoint("TOPRIGHT", status_lbl, "BOTTOMRIGHT", 0, -4)
	bidders2_lbl:SetJustifyH("RIGHT")
	group.bidders2_lbl = bidders2_lbl

	local editbox = AceGUI:Create("EditBox")
	editbox:SetWidth(80)
	editbox:ClearAllPoints()
	editbox:SetUserData('item', self)
	editbox:SetCallback("OnEnterPressed", function(widget, event, value)
		if value == "" then value = nil end
		widget:GetUserData('item').auto_bid = value
	end)
	editbox:SetPoint("TOPRIGHT", frame, "TOPRIGHT", -10, -24)
	group:AddChild(editbox)

	local winners_lbl = frame:CreateFontString()
	winners_lbl:SetFontObject("GameFontHighlight")
	winners_lbl:SetPoint("TOPLEFT", bidders2_lbl, "TOPRIGHT", 16, 0)
	winners_lbl:SetPoint("RIGHT", frame, "RIGHT", -10, 0)
	winners_lbl:SetJustifyH("LEFT")
	group.winners_lbl = winners_lbl

	group.ms_btn = CreateItemButton(frame, "Dice", "Bidding Main Spec", "TOPRIGHT", frame, "TOPRIGHT", -64, -6)
	group.ms_btn.parent = self
	group.ms_btn:SetScript("OnClick", function(frame) frame.parent:SetStatus(STATUS_MAIN_SPEC) end)

	group.os_btn = CreateItemButton(frame, "Coin", "Bidding Off Spec", "TOPRIGHT", frame, "TOPRIGHT", -36, -6)
	group.os_btn.parent = self
	group.os_btn:SetScript("OnClick", function(frame) frame.parent:SetStatus(STATUS_OFF_SPEC) end)

	group.pass_btn = CreateItemButton(frame, "Pass", "Not Bidding", "TOPRIGHT", frame, "TOPRIGHT", -8, -4)
	group.pass_btn.parent = self
	group.pass_btn:SetScript("OnClick", function(frame) frame.parent:SetStatus(STATUS_NOT_BIDDING) end)

	self.group = group
	self:UpdateFrameInitial()
	self:UpdateFrame()
end

local function UnitString(unit)
	if unit then
		local _, classFileName = UnitClass(unit)
		if classFileName then
			return "|c"..(RAID_CLASS_COLORS[classFileName].colorStr)..unit
		else
			return unit
		end
	else
		return ""
	end
end

function LootItem:UpdateAllVisibilities()
	for item_link, loot in pairs(lootList) do
		loot:UpdateVisiblity()
	end
end

function LootItem:UpdateVisiblity()
	if not ARPC.ShowAll and self.status == STATUS_NOT_BIDDING then
		--self.group.frame:Hide()
	else
		--self.group.frame:Show()
	end
end

function LootItem:UpdateFrameInitial()
	local name, link, quality, iLevel, reqLevel, class, subclass, maxStack, equipSlot, texture, vendorPrice = GetItemInfo(self.item_link)

	if link then
		local quantity_txt
		if self.quantity > 1 then
			quantity_txt = "x"..self.quantity
		else
			quantity_txt = ""
		end
		if ilvlMap[ iLevel ] then
			self.group.item_lbl:SetText(link..quantity_txt..GREEN_FONT_COLOR_CODE.." "..ilvlMap[ iLevel ])
		else
			self.group.item_lbl:SetText(link..quantity_txt..YELLOW_FONT_COLOR_CODE.." ("..iLevel..")")
		end
		self.group.icon:SetNormalTexture(texture)
	else
		table.insert(waitingForItemData, self.item_link)
		self.group.item_lbl:SetText(RED_FONT_COLOR_CODE.."Fetching item info...")
	end
end


function LootItem:UpdateFrame()
	if not self.bidList then
		if self.handler then
			self.group.status_lbl:SetText("Bidding Open - "..UnitString(self.handler))
		end
	elseif self.bidList[BIDLIST_State] == STATE_NOT_STARTED then
		self.group.status_lbl:SetText("Bidding Open - "..UnitString(self.handler))
	elseif self.bidList[BIDLIST_State] == STATE_LAST_CALL then
		self.group.status_lbl:SetText("Last Call - "..UnitString(self.handler))
	elseif self.bidList[BIDLIST_State] == STATE_FINISHED then
		local status = GREEN_FONT_COLOR_CODE .. "Finished|r - "
		for index,list in pairs(self.winners) do
			if index > 1 then status = status .. ", " end
			if list[1] then
				status = status .. UnitString(list[1]).." "..list[2].."|r"
			else
				status = status .. GRAY_FONT_COLOR_CODE .. "Disenchanted|r"
			end
		end
		self.group.status_lbl:SetText(status)
	elseif self.bidList[BIDLIST_State] >= STATE_BIDDING_MS and self.bidList[BIDLIST_State] < STATE_FINISHED then
		local status = nil
		local bidder = self.bidList[BIDLIST_HighBidder]
		local bid = self:GetHighBidAmount()
		if bidder then
			status = UnitString(bidder) .. ": " .. bid .. "|r"
		else
			status = "No bids yet"
		end
		if self:CurrentBidder() then
			status = status .. " - " .. UnitString(self:CurrentBidder()) .. "'s bid|r"
		end
		
		self.group.status_lbl:SetText(status)
	end
	
	if self.status == STATUS_MAIN_SPEC then
		self.group.ms_btn:SetAlpha(1)
		self.group.os_btn:SetAlpha(0.3)
		self.group.pass_btn:SetAlpha(0.3)
	elseif self.status == STATUS_OFF_SPEC then
		self.group.ms_btn:SetAlpha(0.3)
		self.group.os_btn:SetAlpha(1)
		self.group.pass_btn:SetAlpha(0.3)
	elseif self.status == STATUS_NOT_BIDDING then
		self.group.ms_btn:SetAlpha(0.3)
		self.group.os_btn:SetAlpha(0.3)
		self.group.pass_btn:SetAlpha(1)
	else
		self.group.ms_btn:SetAlpha(1)
		self.group.os_btn:SetAlpha(1)
		self.group.pass_btn:SetAlpha(1)
	end
end

function LootItem:UpdateBiddersList()
	if not self.bidList or self.bidList[BIDLIST_State] == STATE_FINISHED then
		self.group.bidders_lbl:SetText("")
		self.group.bidders2_lbl:SetText("")
		self.group.winners_lbl:SetText("")
		self.group:SetHeight(52)
		ARPC.window.container:DoLayout()
		return
	end
	local winners_text = ""
	for index,list in pairs(self.winners) do
		if winners_text ~= "" then winners_text = winners_text .. "\n" end
		winners_text = winners_text .. "#" .. index .. ": "
		if list[1] then
			winners_text = winners_text .. UnitString(list[1]).." "..list[2].."|r"
		else
			winners_text = winners_text .. "Disenchanted"
		end
	end
	local text = ""
	local text_right = ""
	local height = 0
	local first_tier = nil
	for index,list in pairs(self.bidList[BIDLIST_Bidders]) do
		for offset,bidder in pairs(list) do
			if text ~= "" then text = text .. "\n" end
			if text_right ~= "" then text_right = text_right .. "\n" end
			if not first_tier then first_tier = index end
			
			local current_tier
			if self.bidList[BIDLIST_State] < STATE_BIDDING_MS then
				current_tier = index == first_tier
			else
				current_tier = (self.bidList[BIDLIST_State] - STATE_BIDDING_MS + 1) == index
			end
			
			local tier_color = nil

			if current_tier then
				text = text .. UnitString(bidder[BIDLIST_Bidders_Player]) .. "|r"	
			else
				text = text .. GRAY_FONT_COLOR_CODE .. (bidder[BIDLIST_Bidders_Player]) .. "|r"
				tier_color = GRAY_FONT_COLOR_CODE
			end
			
			local bid = bidder[BIDLIST_Bidders_Bid]
			if bid == 0 then
				text = text .. (tier_color or "") .. " - " .. (tier_color or YELLOW_FONT_COLOR_CODE) .. "Pass|r"
			elseif bid then
				if bidder[BIDLIST_Bidders_Player] == self.bidList[BIDLIST_HighBidder] then
					text = text .. (tier_color or "") .. " - " .. (tier_color or GREEN_FONT_COLOR_CODE) .. bid .. "|r"
				else
					text = text .. (tier_color or "") .. " - " .. (tier_color or YELLOW_FONT_COLOR_CODE) .. bid .. "|r"
				end
			end
			
			local group = ""
			if index == BIDLIST_Bidders_OS then
				group = "OS "
			elseif index == BIDLIST_Bidders_MSTrial then
				group = "Trial "
			elseif index == BIDLIST_Bidders_OSTrial then
				group = "Trial OS "
			end
			
			text_right = text_right .. (tier_color or "") .. group .. bidder[BIDLIST_Bidders_Points] .. "|r"
			
			height = height + 13
		end
	end
	self.group.bidders_lbl:SetText(text)
	self.group.bidders2_lbl:SetText(text_right)
	self.group.winners_lbl:SetText(winners_text)
	self.group:SetHeight(52 + max(self.group.winners_lbl:GetHeight(), self.group.bidders_lbl:GetHeight()) )
	ARPC.window.container:DoLayout()
end

function RemoveWidget(widget)
	widget:PauseLayout()
	widget:Fire("OnRelease")
	widget:ReleaseChildren()

	if widget.OnRelease then
		widget:OnRelease()
--	else
--		error(("Widget type %s doesn't supply an OnRelease Function"):format(widget.type))
	end
	for k in pairs(widget.userdata) do
		widget.userdata[k] = nil
	end
	for k in pairs(widget.events) do
		widget.events[k] = nil
	end
	widget.width = nil
	widget.relWidth = nil
	widget.height = nil
	widget.relHeight = nil
	widget.noAutoHeight = nil
	widget.frame:ClearAllPoints()
	widget.frame:Hide()
	widget.frame:SetParent(UIParent)
	widget.frame.width = nil
	widget.frame.height = nil
	if widget.content then
		widget.content.width = nil
		widget.content.height = nil
	end
end

function LootItem:ItemFinished()
	if self.bidList then
		self.bidList[BIDLIST_State] = STATE_FINISHED
		self:UpdateFrame()
		self:UpdateBiddersList()
	else
		self:RemoveItem()
	end
end

function LootItem:RemoveItem()
	for index,entry in ipairs(ARPC.window.container.children) do
		if entry == self.group then
			table.remove(ARPC.window.container.children, index)
			break
		end
	end
	RemoveWidget( self.group )
	lootList[ self.item_link ] = nil
	ARPC.window.container:DoLayout()
end

function LootItem:ClearFinished()
	for link,item in pairs( lootList ) do
		if item.bidList and item.bidList[BIDLIST_State] == STATE_FINISHED then
			item:RemoveItem()
		end
	end
end

function LootItem:Error(text)
	Error(YELLOW_FONT_COLOR_CODE.."("..self.item_link..")|r "..text)
end
function LootItem:Alert(text)
	ARPC:Print("("..self.item_link..")|r "..text)
end

function LootItem:CurrentMinimumBid()
	return self.bidList[BIDLIST_MinBid]
end

function LootItem:CurrentBidder()
	if not self.bidList[BIDLIST_CurrentBidder] then return nil end
	local list = nil
	if self.bidList[BIDLIST_State] == STATE_BIDDING_MS then
		list = self.bidList[BIDLIST_Bidders][BIDLIST_Bidders_MS]
	elseif self.bidList[BIDLIST_State] == STATE_BIDDING_OS then
		list = self.bidList[BIDLIST_Bidders][BIDLIST_Bidders_OS]
	elseif self.bidList[BIDLIST_State] == STATE_BIDDING_MSTRIAL then
		list = self.bidList[BIDLIST_Bidders][BIDLIST_Bidders_MSTrial]
	elseif self.bidList[BIDLIST_State] == STATE_BIDDING_OSTRIAL then
		list = self.bidList[BIDLIST_Bidders][BIDLIST_Bidders_OSTrial]
	end
	if not list then return nil end

	return list[ self.bidList[BIDLIST_CurrentBidder ] ][ BIDLIST_Bidders_Player ]
end

function LootItem:ShowBidPopup()
	local popup_name = "ARP_BID_POPUP_"..self.item_link
	local item_link = self.item_link

	if StaticPopupDialogs[popup_name] == nil then

		StaticPopupDialogs[popup_name] = {
			button1 = "Send",
			button2 = CANCEL,
			OnAccept = function(self)
					local text = self.editBox:GetText()
					return not LootItem:get(item_link):EnteredBid(text)
			end,
			EditBoxOnEnterPressed = function(self)
					local text = self:GetParent().editBox:GetText()
					if LootItem:get(item_link):EnteredBid(text) then
						self:GetParent():Hide()
					end
			end,
			hasEditBox = true,
			whileDead = true,
			hideOnEscape = 1,
			preferredIndex = 3
		}
	end

	local link = select(2, GetItemInfo(self.item_link))
	
	local text = "It is your turn to bid on "..link
	local highBidAmt = self:GetHighBidAmount()
	if highBidAmt then
		text = text .. "\n\nThe highest bidder is "..UnitString(self.bidList[BIDLIST_HighBidder]).."|r for "..YELLOW_FONT_COLOR_CODE..highBidAmt.."|r"
		text = text .. "\n\nYour minimum bid is "..YELLOW_FONT_COLOR_CODE..self:CurrentMinimumBid().."|r."
	else
		text = text .. "\n\nThere have been no bids yet.\n\nThe minimum bid is "..YELLOW_FONT_COLOR_CODE..self:CurrentMinimumBid().."|r."
	end
	StaticPopupDialogs[popup_name].text = text

	StaticPopup_Show(popup_name)
end

function LootItem:EnteredBid(text)
	local bid
	local minimum_bid = self:CurrentMinimumBid()

	text = strlower(text)
	if text == "yup" then text = "+1" end
	if text == "nope" then text = "pass" end
	if text ~= 'pass' and not strmatch(text, '^\+?%d+$') then
		Error("You have entered an invalid bid. (" .. text .. ")")
		return false
	elseif text == "pass" then
		self:SendBid(0)
		return true
	elseif strsub(text, 1, 1) == "+" then
		bid = minimum_bid - 1 + tonumber(strsub(text, 2))
	else
		bid = tonumber(text)
	end
	if bid < minimum_bid then
		Error("Your bid of "..bid.." is less the minimum bid of "..minimum_bid..".")
		return false
	elseif not ARPC_CurrentPoints then
		Error("You must enter your current point total with '/arpc points [amount]' before you can continue.")
		return false
	elseif bid > ARPC_CurrentPoints then
		Error("Your bid of "..bid.." exceeds your maximum point total.")
		return false
	else
		self:SendBid(bid)
	end
	return true
end

function LootItem:get(item_link)
	return lootList[ item_link ]
end

function LootItem:add(data)
	local loot = lootList[ data[LOOT_Items_Item_Link] ]
	if loot then
		if loot.bidList and loot.bidList[BIDLIST_State] == STATE_FINISHED then
			loot:RemoveItem()
		else
			loot.quantity = data[LOOT_Items_Quantity]
			loot.handler = data[LOOT_Items_Handler]
		
			loot:UpdateFrameInitial()
			loot:UpdateFrame()
			loot:UpdateBiddersList()
			loot:CheckBidding()
			return false
		end
	end
	loot = {
		item_link = data[LOOT_Items_Item_Link],
		quantity = data[LOOT_Items_Quantity],
		handler = data[LOOT_Items_Handler],
		bidList = nil,
		
		status = nil, -- Status of user bidding
		bidding_sent = nil, -- nil for not sent, false for send but no ack, true for sent and acked
		bid_sent = nil, -- nil for not sent, false for send but no ack, true for sent and acked
		auto_bid = nil,
		winners = {}
	}
	lootList[ loot.item_link ] = loot
	setmetatable(loot, self)
	self.__index = self
	loot:CreateFrame()
	return true
end

function LootItem:GetHighBidAmount()
	local list = nil
	if self.bidList[BIDLIST_State] == STATE_BIDDING_MS then
		list = self.bidList[BIDLIST_Bidders][BIDLIST_Bidders_MS]
	elseif self.bidList[BIDLIST_State] == STATE_BIDDING_OS then
		list = self.bidList[BIDLIST_Bidders][BIDLIST_Bidders_OS]
	elseif self.bidList[BIDLIST_State] == STATE_BIDDING_MSTRIAL then
		list = self.bidList[BIDLIST_Bidders][BIDLIST_Bidders_MSTrial]
	elseif self.bidList[BIDLIST_State] == STATE_BIDDING_OSTRIAL then
		list = self.bidList[BIDLIST_Bidders][BIDLIST_Bidders_OSTrial]
	end
	if list == nil then return nil end

	local highBidder = nil

	for index, entry in pairs( list ) do
		if entry[BIDLIST_Bidders_Bid] and (highBidder == nil or entry[BIDLIST_Bidders_Bid] >= highBidder) then
			highBidder = entry[BIDLIST_Bidders_Bid]
		end
	end

	return highBidder
end

function LootItem:OnCurrentBidder()
	if self.auto_bid then
		local i1, i2 = strmatch(self.auto_bid, "^(%d+) (%d+)$")
		local highBidAmt = self:GetHighBidAmount()
		if self.auto_bid == 'pass' then
			self:EnteredBid('pass')
		elseif i1 and i2 then
			if highBidAmt and highBidAmt >= tonumber(i2) then
				self:ShowBidPopup()
			else
				self:EnteredBid(i1)
			end
			return
		else
			local i3 = strmatch(self.auto_bid, "^(\+?%d+)$")
			if i3 then
				if strsub(i3, 1, 1) == "+" or not highBidAmt or tonumber(i3) > highBidAmt then
					self:EnteredBid(i3)
				else
					self:ShowBidPopup()
				end
				return
			end
		end
		Error("Your auto bid '"..self.auto_bid.."' was in an invalid format.")
	else
		self:ShowBidPopup()
	end
end

function LootItem:SetBidList(data)
	self.bidList = data
	self:UpdateFrame()
	self:UpdateBiddersList()
	self:CheckBidding()
	if self:CurrentBidder() == UnitName('player') then
		self:OnCurrentBidder()
	end
end

function LootItem:CheckBidding()
	if self.handler and (not self.bidList or self.bidList[BIDLIST_State] < STATE_BIDDING_MS) and self.bidding_sent == nil and (self.status == STATUS_OFF_SPEC or self.status == STATUS_MAIN_SPEC) then
		self:SendBidding()
	end
end

function LootItem:SendBid(amount)
	if self.handler and self.bidList and self:CurrentBidder() == UnitName("player") then
		self.bid_sent = false
		ARPC:SendMessage({ [COMMAND]="BID", [BID_Item_Link]=self.item_link, [BID_Bid]=amount }, "WHISPER", self.handler)
	else 
		self:Error("Unable to bid for item")
	end
end

function LootItem:ReceivedBidAck(error)
	if self.bid_sent == false then
		if error then
			self:Error("Bid error: "..error)
			self.bid_sent = nil
		else
			self.bid_sent = true
		end
	else
		self:Error("Received unexpected Bid Ack")
	end
end

function LootItem:SendBidding()
	if not ARPC_CurrentPoints then
		Error("You must enter your current point total with '/arpc points [amount]' before you can continue.")
	elseif self.handler and (not self.bidList or self.bidList[BIDLIST_State] < STATE_BIDDING_MS) then
		if self.status == STATUS_NOT_BIDDING then
			if self.bidding_sent ~= nil then
				ARPC:SendMessage({ [COMMAND]="BIDDING", [BIDDING_Item_Link]=self.item_link, [BIDDING_Points]=nil }, "WHISPER", self.handler)
				self.bidding_sent = false
			end
		else
			ARPC:SendMessage({ [COMMAND]="BIDDING", [BIDDING_Item_Link]=self.item_link, [BIDDING_Points]=ARPC_CurrentPoints, [BIDDING_MS]=(self.status == STATUS_MAIN_SPEC) }, "WHISPER", self.handler)
			self.bidding_sent = false
		end
	else
		self:Error("Couldn't send BIDDING")
	end
end

function LootItem:ReceivedBiddingAck(error)
	if self.bidding_sent == false then
		if error then
			self:Error("Bidding error: "..error)
			self.bidding_sent = nil
		else
			self.bidding_sent = true
		end
	else
		self:Error("Received unexpected Bidding Ack")
	end
end


function LootItem:SetStatus(status)
	if self.bidList and self.bidList[BIDLIST_State] >= STATE_BIDDING_MS then
		Error("Can't set item status once bidding has already started")
		return
	end

	self.status = status
	self:UpdateFrame()
	self:UpdateVisiblity()
	if self.handler == nil then
		-- ARPC:Print("Bidding for this item hasn't started yet. Your interest will be sent when it has.")
	else
		self:SendBidding()
	end
end

function LootItem:ProcessWinner(number, player, bid)
	table.insert(ARPC_LootLog, {
		self.item_link,
		time(),
		player,
		bid
	})
	self.winners[number] = { player, bid }
	if ARPC_CurrentPoints and player == UnitName("player") then
		ARPC_CurrentPoints = ARPC_CurrentPoints - bid
		UpdateStatusBar()
	end
	
	self:UpdateFrame()
	self:UpdateBiddersList()
end

function ARPC:OnInitialize()
	-- Log of outcomes of previous items with entries consisting of { item_link, timestamp, player (nil for DE), bid }
	if ARPC_LootLog == nil then ARPC_LootLog = {} end
	
	if (ARPC_CurrentPoints == nil) then
		ARPC_CurrentPoints = 0
		Error("Your current point total is not set. You must set it with /arpc points [amount], before you can continue.")
	end
	
	local options = {
		name = "ARPC",
		handler = ARPC,
		type = "group",
		args = {
			toggle = {
				type = "execute",
				name = "Toggle Window",
				desc = "Shows/hides the main window (also available in game keybindings)",
				func = function() ARPC_Toggle() end
			},
			test = {
				type = "execute",
				name = "Test Items",
				desc = "Loads some test items in",
				func = function() TestItems() end
			},
			test2 = {
				type = "execute",
				name = "Test Items",
				desc = "Loads some test items in",
				func = function() TestItems2() end
			},
			points = {
				type = "range",
				name = "Points",
				desc = "Sets/gets your current point total",
				step = 1,
				bigStep = 10,
				min = -99999,
				max = 99999,
				softMin = 0,
				softMax = 3000,
				get = function(info) return ARPC_CurrentPoints end,
				set = function(info, val)
					ARPC_CurrentPoints = val
					for link,item in pairs( lootList ) do
						item:CheckBidding()
					end
					UpdateStatusBar()
				end
			}
		}
	}

	LibStub("AceConfig-3.0"):RegisterOptionsTable("Angry Raid Points Client", options, {"arpc"})
end

function ARPC:OnEnable()
	ARPC:RegisterEvent("GET_ITEM_INFO_RECEIVED")

	ARPC:RegisterComm("ARP", "ReceiveMessage")
	
	ARPC:ScheduleTimer("AfterEnable", 5)

	CreateWindow()
end

function ARPC:AfterEnable()
	ARPC:SendMessage({ "REQUEST_STATUS" })
end
